/*
Copyright 2020 The Flux authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package dependency

import (
	"fmt"
	"slices"
	"strings"

	"sigs.k8s.io/controller-runtime/pkg/client"

	"github.com/fluxcd/pkg/apis/meta"
)

// Dependent interface defines methods that a Kubernetes resource object should
// implement in order to use the dependency package for ordering dependencies.
type Dependent interface {
	client.Object
	meta.ObjectWithDependencies
}

const (
	unmarked = iota
	permanentMark
	temporaryMark
)

// Sort takes a slice of Dependent objects and returns a sorted slice of
// NamespacedObjectReference based on their dependencies. It performs a
// topological sort using a depth-first search algorithm, which has
// runtime complexity of O(|V| + |E|), where |V| is the number of
// vertices (objects) and |E| is the number of edges (dependencies).
//
// Reference:
// https://en.wikipedia.org/wiki/Topological_sorting#Depth-first_search
func Sort(V []Dependent) ([]meta.NamespacedObjectReference, error) {
	sorted := make([]meta.NamespacedObjectReference, 0, len(V))

	// Compute the vertex and reference maps for each object.
	vertex := make(map[meta.NamespacedObjectReference]int, len(V))
	reference := make([]meta.NamespacedObjectReference, len(V))
	mark := make([]int, len(V))
	for u, obj := range V {
		key := meta.NamespacedObjectReference{
			Name:      obj.GetName(),
			Namespace: obj.GetNamespace(),
		}
		vertex[key] = u
		reference[u] = key
	}

	// Compute topological order of objects using a depth-first search.
	var depthFirstSearch func(u int) (cycle []string)
	depthFirstSearch = func(u int) []string {
		mark[u] = temporaryMark
		for _, key := range V[u].GetDependsOn() {
			v := vertex[key]
			if mark[v] == permanentMark {
				continue
			}
			if mark[v] == temporaryMark {
				// Cycle detected.
				return []string{key.String(), reference[u].String()}
			}
			if cycle := depthFirstSearch(v); len(cycle) > 0 {
				return append(cycle, reference[u].String())
			}
		}
		mark[u] = permanentMark
		sorted = append(sorted, reference[u])
		return nil
	}
	for u := range V {
		if mark[u] == unmarked {
			if cycle := depthFirstSearch(u); len(cycle) > 0 {
				slices.Reverse(cycle)
				return nil, fmt.Errorf("circular dependency detected: %v", strings.Join(cycle, " -> "))
			}
		}
	}

	return sorted, nil
}
