/*
Copyright 2020 The Flux authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package dependency_test

import (
	"testing"

	. "github.com/onsi/gomega"
	corev1 "k8s.io/api/core/v1"
	v1 "k8s.io/apimachinery/pkg/apis/meta/v1"

	"github.com/fluxcd/pkg/apis/meta"
	"github.com/fluxcd/pkg/runtime/dependency"
)

type mockDependent struct {
	corev1.Node
	dependsOn []meta.NamespacedObjectReference
}

func (d mockDependent) GetDependsOn() []meta.NamespacedObjectReference {
	return d.dependsOn
}

func TestSort(t *testing.T) {
	for _, tt := range []struct {
		name string
		V    []dependency.Dependent
		want []meta.NamespacedObjectReference
		err  string
	}{
		{
			name: "simple",
			V: []dependency.Dependent{
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "default",
							Name:      "frontend",
						},
					},
					dependsOn: []meta.NamespacedObjectReference{
						{
							Namespace: "linkerd",
							Name:      "linkerd",
						},
						{
							Namespace: "default",
							Name:      "backend",
						},
					},
				},
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "linkerd",
							Name:      "linkerd",
						},
					},
				},
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "default",
							Name:      "backend",
						},
					},
					dependsOn: []meta.NamespacedObjectReference{
						{
							Namespace: "linkerd",
							Name:      "linkerd",
						},
					},
				},
			},
			want: []meta.NamespacedObjectReference{
				{
					Namespace: "linkerd",
					Name:      "linkerd",
				},
				{
					Namespace: "default",
					Name:      "backend",
				},
				{
					Namespace: "default",
					Name:      "frontend",
				},
			},
		},
		{
			name: "circular dependency",
			V: []dependency.Dependent{
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "default",
							Name:      "dependency",
						},
					},
					dependsOn: []meta.NamespacedObjectReference{
						{
							Namespace: "default",
							Name:      "endless",
						},
					},
				},
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "default",
							Name:      "endless",
						},
					},
					dependsOn: []meta.NamespacedObjectReference{
						{
							Namespace: "default",
							Name:      "circular",
						},
					},
				},
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "default",
							Name:      "circular",
						},
					},
					dependsOn: []meta.NamespacedObjectReference{
						{
							Namespace: "default",
							Name:      "dependency",
						},
					},
				},
			},
			err: "circular dependency detected: default/dependency -> default/endless -> default/circular -> default/dependency",
		},
		{
			name: "missing namespace",
			V: []dependency.Dependent{
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "application",
							Name:      "backend",
						},
					},
				},
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "application",
							Name:      "frontend",
						},
					},
					dependsOn: []meta.NamespacedObjectReference{
						{
							Name: "backend",
						},
					},
				},
			},
			want: []meta.NamespacedObjectReference{
				{
					Namespace: "application",
					Name:      "backend",
				},
				{
					Namespace: "application",
					Name:      "frontend",
				},
			},
		},
		{
			name: "dead end",
			V: []dependency.Dependent{
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "default",
							Name:      "backend",
						},
					},
					dependsOn: []meta.NamespacedObjectReference{
						{
							Namespace: "default",
							Name:      "common",
						},
					},
				},
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "default",
							Name:      "frontend",
						},
					},
					dependsOn: []meta.NamespacedObjectReference{
						{
							Namespace: "default",
							Name:      "infra",
						},
					},
				},
				&mockDependent{
					Node: corev1.Node{
						ObjectMeta: v1.ObjectMeta{
							Namespace: "default",
							Name:      "common",
						},
					},
				},
			},
			want: []meta.NamespacedObjectReference{
				{Name: "common", Namespace: "default"},
				{Name: "backend", Namespace: "default"},
				{Name: "frontend", Namespace: "default"},
			},
		},
	} {
		t.Run(tt.name, func(t *testing.T) {
			g := NewWithT(t)

			got, err := dependency.Sort(tt.V)

			if tt.err != "" {
				g.Expect(err).To(HaveOccurred())
				g.Expect(err.Error()).To(Equal(tt.err))
				g.Expect(got).To(BeNil())
			} else {
				g.Expect(err).NotTo(HaveOccurred())
				g.Expect(got).To(Equal(tt.want))
			}
		})
	}
}
